#!/usr/bin/perl -w

use strict;
use warnings;
use FileHandle;
use Getopt::Long;

use constant true => 1;
use constant false => 0;
use constant DEFAULT_IDENTITY => 1;
use constant DEFAULT_COVERAGE => 1;
use constant DEFAULT_EPCR_SIZE => 0;

# Given a list of numbers, returns the minimum value in the list
sub min() {
	my @sorted = sort { $a <=> $b } @_;
	return shift(@sorted);
} # End of sub min()

# Given a list of numbers, returns the maximum value in the list 
sub max() {
	my @sorted = sort { $a <=> $b } @_;
	return pop(@sorted);
} # End of sub max


# -------------------- Main Program Starts Here --------------------
my ($lengthFile, $blastoutputFile, $pairsFile, $identOpt, $coverageOpt, $maxPCROpt);

my $options = &GetOptions("length|l=s" => \$lengthFile,
                          "blast|b=s" => \$blastoutputFile,
						  "pairs|p=s" => \$pairsFile,
						  "ident|i:f" => \$identOpt,
						  "coverage|c:f" => \$coverageOpt,
						  "maxpcr|e:i" => \$maxPCROpt);

if (defined($lengthFile) && defined($blastoutputFile) && defined($pairsFile)) {
	# Get default values
	$identOpt = DEFAULT_IDENTITY if (!defined($identOpt));
	$coverageOpt = DEFAULT_COVERAGE if (!defined($coverageOpt));
	$maxPCROpt = DEFAULT_EPCR_SIZE if (!defined($maxPCROpt));

	my %primers;
	my $fh = new FileHandle();
	open ($fh, $lengthFile) or die("Cannot open primer length file");
	while (<$fh>) {
		chomp;
		my ($name, $length) = split(/\t/);
		$primers{$name}->{"length"} = $length;
	} # End of while loop
	close ($fh);

	# From the blast hits (-m 8 output, filter all the primer hits
	# with 100% identity and 100% coverage
	open ($fh, $blastoutputFile) or die("Cannot open blast m 8 output file");
	while (<$fh>) {
		chomp;
		if ($_ !~ m/^Query ID/) {
			my ($query, $subject, $ident, $alignLength, $mismatch, $gaps, $qstart, $qend,
				$sstart, $send, $eval, $score) = split(/\t/, $_);

			if ($subject =~ m/^gi\|(\d+)\|gb\|(\S+)\|/) {
				$subject = $2;
			} # End of if statemen

			if (exists $primers{$query}) {
				# check coverage and identity
				if ($ident >= ($identOpt * 100) && ($alignLength / $primers{$query}->{"length"}) >= $coverageOpt) {
					my $strand = "+";
					$strand = "-" if ($sstart > $send);
					my $hit = sprintf("%s,%s,%2.1f%%,%2.1f%%,%s,%s", $subject, $strand, $ident, 
					                  ($alignLength / $primers{$query}->{"length"}) * 100, &min($sstart, $send), &max($sstart, $send));
					if (exists $primers{$query}->{"hits"}) {
						$primers{$query}->{"hits"} = join(";", $primers{$query}->{"hits"}, $hit);
					} # End of if statmeent
					else {
						$primers{$query}->{"hits"} = $hit;
					} # End of else statement
				} # End of if statmeent
			} # End of if statement
		} # End of if statement
	} # End of while loop
	close ($fh);

	# Checking primer pairs for their direction and amplicon size
	open ($fh, $pairsFile) or die("Cannot open primer pairs file");
	printf("Forward Name\tForward Ident. %%\tForward Cvg. %%\tForward Start\tForward End\tReverse\tReverse Ident. %%\tReverse Cvg. %%\tReverse Start\tReverse End\tHit\tAmplicon Start\tAmplicon End\tExpected Product Size\tePCR Product Size\tProduct Difference\n");

	while (<$fh>) {
		chomp;
		my ($p1, $p2, $amplicon) = split(/\t/, $_);

		if (exists $primers{$p1}->{"hits"} && exists $primers{$p2}->{"hits"}) {
			my @p1hits = split(/;/, $primers{$p1}->{"hits"});
			my @p2hits = split(/;/, $primers{$p2}->{"hits"});
			my $printed = false; 

			for (my $i=0; $i < scalar(@p1hits); $i++) {
				my ($p1name, $p1strand, $p1ident, $p1coverage, $p1start, $p1end) = split(/,/, $p1hits[$i]);

				for (my $j=0; $j < scalar(@p2hits); $j++) {
					my ($p2name, $p2strand, $p2ident, $p2coverage, $p2start, $p2end) = split(/,/, $p2hits[$j]);

					# Check for same name and different strand
					if ($p1name eq $p2name && $p1strand ne $p2strand) {
						my ($start, $end);
						if ($p1strand eq "+") {
							$start = $p1start;
							$end = $p2end;
						} # End of if stamtent
						else {
							$start = $p2start;
							$end = $p1end;
						} # End of else statement

						my $epcr = $end - $start + 1;
						
						if ($epcr >= 0) {
							if ($maxPCROpt == 0 || $epcr <= $maxPCROpt) {
								if ($p1name =~ m/(\S+)_(\d+)-(\d+)$/) {
									printf("%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t", $p1, $p1ident, $p1coverage, $p1start + $2 - 1, 
									       $p1end + $2 - 1, $p2, $p2ident, $p2coverage, $p2start + $2 - 1, $p2end + $2 - 1, $1);
									my $diff;

									if ($amplicon != 0) {
										$diff = &max($amplicon, $epcr) - &min($amplicon, $epcr);
									} # End of if statement
									else {
										$amplicon = "-";
										$diff = "-";
									} # End of if statement
									printf("%s\t%s\t%s\t%s\t%s\n", $start + $2 - 1, $end + $2 - 1, $amplicon, $epcr, $diff);
								} # End of if statement
								else {
									printf("%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t", $p1, $p1ident, $p1coverage, $p1start, $p1end, 
								    	   $p2, $p2ident, $p2coverage, $p2start, $p2end, $p1name);
									my $diff;
									if ($amplicon != 0) {
										$diff = &max($amplicon, $epcr) - &min($amplicon, $epcr);
									} # End of if statement
									else {
										$amplicon = "-";
										$diff = "-";
									} # End of if statement
									printf("%s\t%s\t%s\t%s\t%s\n", $start, $end, $amplicon, $epcr, $diff);
								} # End of if statement
								
								$printed = true;
							} # End of if statmeent
						} # End of if statmenet
					} # End of if statement
				} # End of for loop
			} # End of for loop
		
			if ($printed == false) {
				# both primers have hits but no amplicon with the specified criterias could
				# be determined 
				printf("%s\t-\t-\t-\t-\t%s\t-\t-\t-\t-\t-\t-\t-\t%s\t-\t-\n", $p1, $p2, $amplicon);
			} # End of if statement
		} # end of if statement
		else {		# this else statement takes care of missing p1 or p2 hits
			printf("%s\t-\t-\t-\t-\t%s\t-\t-\t-\t-\t-\t-\t-\t%s\t-\t-\n", $p1, $p2, $amplicon);
		} # End of else statement
	} # End of while loop
	close ($fh);
} # End of if statmenet
else {
	print << "EOFERR";

*** INCORRECT NUMBER OF ARGUMENTS ***
USAGE: perl $0 --length <length file> --blast <tabular blast output file> --pairs <primer pairs file> [OPTIONS]
WHERE: <length file>               : Two column text file separated by tab of which the first column contains
                                     the primer name and the second column contains the primer sequence length
                                     in base pairs.
       <tabular blast output file> : Blast output file of primer sequences blasted (BLASTN) to the reference
                                     sequence with -m 8 output option
       <primer pairs file>         : 2 or 3 column text file separated by tab. The first two columns are the
                                     names of the primers of a given pair. The third column is optional and if
                                     included it should contain the expected product size in base pairs
[OPTIONS]
	--ident or -i                  : Minimum identity of each primer hit from 0 to 1 (DEFAULT: 1)
	--coverage or -c               : Minimum coverage of each primer hit from 0 to 1 (DEFAULT: 1)
	--maxpcr or -e                 : Integer value to filter all the amplicons that are less than --maxpcr

EOFERR
} # End of else statement
